import struct

from .attributes import Attribute
from .constants import Classref, ConstantPool
from .fields import Field
from .methods import Method


# From: http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-4.html

##########################################################################
# Chapter 4: The class File Format
##########################################################################

# This chapter describes the Java Virtual Machine class file format. Each class
# file contains the definition of a single class or interface. Although a class
# or interface need not have an external representation literally contained in a
# file (for instance, because the class is generated by a class loader), we will
# colloquially refer to any valid representation of a class or interface as
# being in the class file format.

# A class file consists of a stream of 8-bit bytes. All 16-bit, 32-bit, and
# 64-bit quantities are constructed by reading in two, four, and eight
# consecutive 8-bit bytes, respectively. Multibyte data items are always stored
# in big-endian order, where the high bytes come first. In the Java SE platform,
# this format is supported by interfaces java.io.DataInput and
# java.io.DataOutput and classes such as java.io.DataInputStream and
# java.io.DataOutputStream.

# This chapter defines its own set of data types representing class file data:
# The types u1, u2, and u4 represent an unsigned one-, two-, or four-byte
# quantity, respectively. In the Java SE platform, these types may be read by
# methods such as readUnsignedByte, readUnsignedShort, and readInt of the
# interface java.io.DataInput.

# This chapter presents the class file format using pseudostructures written in
# a C-like structure notation. To avoid confusion with the fields of classes and
# class instances, etc., the contents of the structures describing the class
# file format are referred to as items. Successive items are stored in the class
# file sequentially, without padding or alignment.

# Tables, consisting of zero or more variable-sized items, are used in several
# class file structures. Although we use C-like array syntax to refer to table
# items, the fact that tables are streams of varying-sized structures means that
# it is not possible to translate a table index directly to a byte offset into
# the table.

# Where we refer to a data structure as an array, it consists of zero or more
# contiguous fixed-sized items and can be indexed like an array.

# Reference to an ASCII character in this chapter should be interpreted to mean
# the Unicode code point corresponding to the ASCII character.


class ClassFileWriter:
    def __init__(self, outfile, constant_pool):
        self._outfile = outfile
        self.constant_pool = constant_pool

    def write_bytes(self, b):
        self._outfile.write(b)

    def write_s1(self, u1):
        self._outfile.write(struct.pack('b', u1))

    def write_u1(self, u1):
        self._outfile.write(struct.pack('B', u1))

    def write_s2(self, u2):
        self._outfile.write(struct.pack('>h', u2))

    def write_u2(self, u2):
        self._outfile.write(struct.pack('>H', u2))

    def write_s4(self, u4):
        self._outfile.write(struct.pack('>i', u4))

    def write_u4(self, u4):
        self._outfile.write(struct.pack('>I', u4))

    def write_f(self, f):
        self._outfile.write(struct.pack('>f', f))

    def write_s8(self, u8):
        self._outfile.write(struct.pack('>q', u8))

    def write_u8(self, u8):
        self._outfile.write(struct.pack('>Q', u8))

    def write_d(self, d):
        self._outfile.write(struct.pack('>d', d))


class ClassFileReader:
    def __init__(self, infile, constant_pool, debug=None):
        self._infile = infile
        self.constant_pool = constant_pool
        if debug is None:
            self.debug = lambda *msg: print(*msg)
        else:
            self.debug = lambda *msg: debug.write(' '.join(str(m) for m in msg) + '\n')

    def read_bytes(self, count):
        return self._infile.read(count)

    def read_u1(self):
        return struct.unpack('B', self._infile.read(1))[0]

    def read_s2(self):
        return struct.unpack('>h', self._infile.read(2))[0]

    def read_u2(self):
        return struct.unpack('>H', self._infile.read(2))[0]

    def read_u4(self):
        return struct.unpack('>I', self._infile.read(4))[0]

    def read_f(self):
        return struct.unpack('>f', self._infile.read(4))[0]

    def read_u8(self):
        return struct.unpack('>Q', self._infile.read(8))[0]

    def read_d(self):
        return struct.unpack('>d', self._infile.read(8))[0]


# ------------------------------------------------------------------------
# 4.1. The ClassFile Structure
# ------------------------------------------------------------------------

# A class file consists of a single ClassFile structure:

class BaseClass:
    # u4             magic;
    # u2             minor_version;
    # u2             major_version;
    # u2             constant_pool_count;
    # cp_info        constant_pool[constant_pool_count-1];
    # u2             access_flags;
    # u2             this_class;
    # u2             super_class;
    # u2             interfaces_count;
    # u2             interfaces[interfaces_count];
    # u2             fields_count;
    # field_info     fields[fields_count];
    # u2             methods_count;
    # method_info    methods[methods_count];
    # u2             attributes_count;
    # attribute_info attributes[attributes_count];

    # Table 4.1. Class access and property modifiers

    ACC_PUBLIC = 0x0001  # Declared public; may be accessed from outside its package.
    ACC_FINAL = 0x0010  # Declared final; no subclasses allowed.
    ACC_SUPER = 0x0020  # Treat superclass methods specially when invoked by the invokespecial instruction.
    ACC_INTERFACE = 0x0200  # Is an interface, not a class.
    ACC_ABSTRACT = 0x0400  # Declared abstract; must not be instantiated.
    ACC_SYNTHETIC = 0x1000  # Declared synthetic; not present in the source code.
    ACC_ANNOTATION = 0x2000  # Declared as an annotation type.
    ACC_ENUM = 0x4000  # Declared as an enum type.

    def __init__(
                self, name, extends=None,
                public=True, final=False, interface=False,
                abstract=False, synthetic=False, annotation=False, enum=False,
                implements=None
            ):

        # Constructor properties.
        self.public = public
        self.final = final
        self.super = True
        self.interface = interface
        self.abstract = abstract
        self.synthetic = synthetic
        self.annotation = annotation
        self.enum = enum

        # The magic item supplies the magic number identifying the class file
        # format; it has the value 0xCAFEBABE.
        self.magic = 0xCAFEBABE

        # The values of the minor_version and major_version items are the minor
        # and major version numbers of this class file. Together, a major and a
        # minor version number determine the version of the class file format.
        # If a class file has major version number M and minor version number m,
        # we denote the version of its class file format as M.m. Thus, class
        # file format versions may be ordered lexicographically, for example,
        # 1.5 < 2.0 < 2.1.

        # A Java Virtual Machine implementation can support a class file format
        # of version v if and only if v lies in some contiguous range Mi.0 ≤ v ≤
        # Mj.m. The release level of the Java SE platform to which a Java
        # Virtual Machine implementation conforms is responsible for determining
        # the range.

        # Oracle's Java Virtual Machine implementation in JDK release 1.0.2
        # supports class file format versions 45.0 through 45.3 inclusive. JDK
        # releases 1.1.* support class file format versions in the range 45.0
        # through 45.65535 inclusive. For k ≥ 2, JDK release 1.k supports class
        # file format versions in the range 45.0 through 44+k.0 inclusive.
        #
        # i.e.,
        #   J2SE 8 = 52 (0x34 hex),
        #   J2SE 7 = 51 (0x33 hex),
        #   J2SE 6.0 = 50 (0x32 hex),
        #   J2SE 5.0 = 49 (0x31 hex),
        #   JDK 1.4 = 48 (0x30 hex),
        #   JDK 1.3 = 47 (0x2F hex),
        #   JDK 1.2 = 46 (0x2E hex),
        #   JDK 1.1 = 45 (0x2D hex).
        #
        # If major_version is set to 51 or higher, and you're using a Java 7
        # virtual machine, you'll need to use the ``-XX:-UseSplitVerifier``
        # on the command line. This disables the use of the StackMapFrame
        # verifier. If you're using a Java 8 or higher VM, you'll need to
        # use -noverify, because StackMapFrame are no longer optional.
        self.major_version = 50
        self.minor_version = 0

        # Each value in the interfaces array must be a valid index into the
        # constant_pool table. The constant_pool entry at each value of interfaces[i],
        # where 0 ≤ i < interfaces_count, must be a CONSTANT_Class_info structure
        # (§4.4.1) representing an interface that is a direct superinterface of this
        # class or interface type, in the left-to-right order given in the source for
        # the type.
        if implements is None:
            self.interfaces = []
        else:
            self.interfaces = [Classref(iface) for iface in implements]

        # Each value in the fields table must be a field_info (§4.5) structure giving a
        # complete description of a field in this class or interface. The fields table
        # includes only those fields that are declared by this class or interface. It
        # does not include items representing fields that are inherited from
        # superclasses or superinterfaces.
        self.fields = []

        # Each value in the methods table must be a method_info (§4.6) structure giving
        # a complete description of a method in this class or interface. If neither of
        # the ACC_NATIVE and ACC_ABSTRACT flags are set in the access_flags item of a
        # method_info structure, the Java Virtual Machine instructions implementing the
        # method are also supplied.

        # The method_info structures represent all methods declared by this class or
        # interface type, including instance methods, class methods, instance
        # initialization methods (§2.9), and any class or interface initialization
        # method (§2.9). The methods table does not include items representing methods
        # that are inherited from superclasses or superinterfaces.
        self.methods = []

        # Each value of the attributes table must be an attribute_info (§4.7) structure.

        # The attributes defined by this specification as appearing in the attributes
        # table of a ClassFile structure are the InnerClasses (§4.7.6), EnclosingMethod
        # (§4.7.7), Synthetic (§4.7.8), Signature (§4.7.9), SourceFile (§4.7.10),
        # SourceDebugExtension (§4.7.11), Deprecated (§4.7.15),
        # RuntimeVisibleAnnotations (§4.7.16), RuntimeInvisibleAnnotations (§4.7.17),
        # and BootstrapMethods (§4.7.21) attributes.

        # If a Java Virtual Machine implementation recognizes class files whose version
        # number is 49.0 or above, it must recognize and correctly read Signature
        # (§4.7.9), RuntimeVisibleAnnotations (§4.7.16), and RuntimeInvisibleAnnotations
        # (§4.7.17) attributes found in the attributes table of a ClassFile structure of
        # a class file whose version number is 49.0 or above.

        # If a Java Virtual Machine implementation recognizes class files whose version
        # number is 51.0 or above, it must recognize and correctly read BootstrapMethods
        # (§4.7.21) attributes found in the attributes table of a ClassFile structure of
        # a class file whose version number is 51.0 or above.

        # A Java Virtual Machine implementation is required to silently ignore any or
        # all attributes in the attributes table of a ClassFile structure that it does
        # not recognize. Attributes not defined in this specification are not allowed to
        # affect the semantics of the class file, but only to provide additional
        # descriptive information (§4.7.1).
        self.attributes = []

        # The value of the this_class item must be a valid index into the constant_pool
        # table. The constant_pool entry at that index must be a CONSTANT_Class_info
        # structure (§4.4.1) representing the class or interface defined by this class
        # file.
        self.this_class = Classref(name)

        # For a class, the value of the super_class item either must be zero or must be
        # a valid index into the constant_pool table. If the value of the super_class
        # item is nonzero, the constant_pool entry at that index must be a
        # CONSTANT_Class_info structure (§4.4.1) representing the direct superclass of
        # the class defined by this class file. Neither the direct superclass nor any of
        # its superclasses may have the ACC_FINAL flag set in the access_flags item of
        # its ClassFile structure.

        # If the value of the super_class item is zero, then this class file must
        # represent the class Object, the only class or interface without a direct
        # superclass.

        # For an interface, the value of the super_class item must always be a valid
        # index into the constant_pool table. The constant_pool entry at that index must
        # be a CONSTANT_Class_info structure representing the class Object.
        self.super_class = Classref(extends if extends else 'java/lang/Object')

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, self.this_class.name)

    @staticmethod
    def read(infile, debug=None):
        constant_pool = ConstantPool()
        reader = ClassFileReader(infile, constant_pool, debug=debug)
        dump = 0 if debug else None

        magic = reader.read_u4()
        minor_version = reader.read_u2()
        major_version = reader.read_u2()

        if dump is not None:
            reader.debug("    " * dump, 'Magic: %x' % (magic))
            reader.debug("    " * dump, 'Version: %s.%s' % (major_version, minor_version))

        reader.constant_pool.read(reader, dump)

        access_flags = reader.read_u2()
        val = reader.read_u2()
        this_class = reader.constant_pool[val].name.bytes.decode('mutf-8')
        val = reader.read_u2()
        super_class = reader.constant_pool[val].name.bytes.decode('mutf-8')

        if dump is not None:
            reader.debug("    " * dump, 'Class %s' % this_class)
            reader.debug("    " * dump, '    Extends %s' % super_class)

            access_description = ', '.join(f for f in [
                    flag if access_flags & mask else None
                    for flag, mask in [
                        ('public', Class.ACC_PUBLIC),
                        ('final', Class.ACC_FINAL),
                        ('super', Class.ACC_SUPER),
                        ('interface', Class.ACC_INTERFACE),
                        ('abstract', Class.ACC_ABSTRACT),
                        ('synthetic', Class.ACC_SYNTHETIC),
                        ('annotation', Class.ACC_ANNOTATION),
                        ('enum', Class.ACC_ENUM),
                    ]
                ] if f)
            reader.debug("    " * dump, '    Flags: 0x%04x%s' % (access_flags, ' (%s)') % (
                access_description if access_description else ''
            ))

        interfaces_count = reader.read_u2()
        if dump is not None:
            reader.debug("    " * (dump + 1), 'Interfaces: (%s)' % interfaces_count)
        for i in range(0, interfaces_count):
            val = reader.read_u2()
            interface = reader.constant_pool[val]
            reader.debug("    " * (dump + 2), interface.name)

        fields_count = reader.read_u2()
        if dump is not None:
            reader.debug("    " * (dump + 1), 'Fields: (%s)' % fields_count)
        for i in range(0, fields_count):
            Field.read(reader, dump=dump + 2 if dump is not None else dump)

        methods_count = reader.read_u2()
        if dump is not None:
            reader.debug("    " * (dump + 1), 'Methods: (%s)' % methods_count)
        for i in range(0, methods_count):
            Method.read(reader, dump=dump + 2 if dump is not None else dump)

        attributes_count = reader.read_u2()
        if dump is not None:
            reader.debug("    " * (dump + 1), 'Attributes: (%s)' % attributes_count)
        for i in range(0, attributes_count):
            Attribute.read(reader, dump=dump + 2 if dump is not None else dump)

        if access_flags & BaseClass.ACC_ENUM:
            klass = Enum
        elif access_flags & BaseClass.ACC_INTERFACE:
            klass = Interface
        else:
            klass = Class

        return klass(
            this_class,
            extends=super_class,
            public=bool(access_flags & BaseClass.ACC_PUBLIC),
            final=bool(access_flags & BaseClass.ACC_FINAL),
            abstract=bool(access_flags & BaseClass.ACC_ABSTRACT),
            # synthetic=bool(access_flags & self.ACC_SYNTHETIC),
            # annotation=bool(access_flags & self.ACC_ANNOTATION),
        )

    def write(self, outfile):
        """Output the classfile to the output stream provided"""
        # First, create a writer.
        constant_pool = ConstantPool()
        writer = ClassFileWriter(outfile, constant_pool)

        self.this_class.resolve(constant_pool)
        self.super_class.resolve(constant_pool)

        for interface in self.interfaces:
            interface.resolve(constant_pool)

        for field in self.fields:
            field.resolve(constant_pool)

        for method in self.methods:
            method.resolve(constant_pool)

        for attribute in self.attributes:
            attribute.resolve(constant_pool)

        # import pprint
        # pprint.pprint(constant_pool._constants)
        # pprint.pprint(constant_pool._constant_pool)

        # Now, write everything out.
        writer.write_u4(self.magic)
        writer.write_u2(self.minor_version)
        writer.write_u2(self.major_version)

        constant_pool.write(writer)

        writer.write_u2(self.access_flags)
        writer.write_u2(writer.constant_pool.index(self.this_class))
        writer.write_u2(writer.constant_pool.index(self.super_class))

        writer.write_u2(self.interfaces_count)
        for interface in self.interfaces:
            writer.write_u2(writer.constant_pool.index(interface))

        writer.write_u2(self.fields_count)
        for field in self.fields:
            field.write(writer)

        writer.write_u2(self.methods_count)
        for method in self.methods:
            method.write(writer)

        writer.write_u2(self.attributes_count)
        for attribute in self.attributes:
            attribute.write(writer)

    @property
    def access_flags(self):
        """The value of the access_flags item is a mask of flags used to denote access
        permissions to and properties of this class or interface. The interpretation
        of each flag, when set, is as shown in Table 4.1.

        A class may be marked with the ACC_SYNTHETIC flag to indicate that it was
        generated by a compiler and does not appear in source code.

        The ACC_ENUM flag indicates that this class or its superclass is declared as
        an enumerated type.

        An interface is distinguished by its ACC_INTERFACE flag being set. If its
        ACC_INTERFACE flag is not set, this class file defines a class, not an
        interface.

        If the ACC_INTERFACE flag of this class file is set, its ACC_ABSTRACT flag
        must also be set (JLS §9.1.1.1). Such a class file must not have its
        ACC_FINAL, ACC_SUPER or ACC_ENUM flags set.

        An annotation type must have its ACC_ANNOTATION flag set. If the
        ACC_ANNOTATION flag is set, the ACC_INTERFACE flag must be set as well. If the
        ACC_INTERFACE flag of this class file is not set, it may have any of the other
        flags in Table 4.1 set, except the ACC_ANNOTATION flag. However, such a class
        file cannot have both its ACC_FINAL and ACC_ABSTRACT flags set (JLS §8.1.1.2).

        The ACC_SUPER flag indicates which of two alternative semantics is to be
        expressed by the invokespecial instruction (§invokespecial) if it appears in
        this class. Compilers to the instruction set of the Java Virtual Machine
        should set the ACC_SUPER flag.

        The ACC_SUPER flag exists for backward compatibility with code compiled by
        older compilers for the Java programming language. In Oracle’s JDK prior to
        release 1.0.2, the compiler generated ClassFile access_flags in which the flag
        now representing ACC_SUPER had no assigned meaning, and Oracle's Java Virtual
        Machine implementation ignored the flag if it was set.

        All bits of the access_flags item not assigned in Table 4.1 are reserved for
        future use. They should be set to zero in generated class files and should be
        ignored by Java Virtual Machine implementations.
        """
        return (
            (self.ACC_PUBLIC if self.public else 0) |
            (self.ACC_FINAL if self.final else 0) |
            (self.ACC_SUPER if self.super else 0) |
            (self.ACC_INTERFACE if self.interface else 0) |
            (self.ACC_ABSTRACT if self.abstract else 0) |
            (self.ACC_SYNTHETIC if self.synthetic else 0) |
            (self.ACC_ANNOTATION if self.annotation else 0) |
            (self.ACC_ENUM if self.enum else 0)
        )

    @property
    def interfaces_count(self):
        """The value of the interfaces_count item gives the number of direct
        superinterfaces of this class or interface type.
        """
        return len(self.interfaces)

    @property
    def fields_count(self):
        """The value of the fields_count item gives the number of field_info structures
        in the fields table. The field_info structures (§4.5) represent all fields,
        both class variables and instance variables, declared by this class or
        interface type.
        """
        return len(self.fields)

    @property
    def methods_count(self):
        """The value of the methods_count item gives the number of method_info structures
        in the methods table.
        """
        return len(self.methods)

    @property
    def attributes_count(self):
        """The value of the attributes_count item gives the number of attributes (§4.7)
        in the attributes table of this class.
        """
        return len(self.attributes)


class Class(BaseClass):
    def __init__(self, name, extends=None, implements=None, public=True, final=False, abstract=False):
        super(Class, self).__init__(
            name, extends, implements=implements, public=public, final=final, abstract=abstract
        )


class Interface(BaseClass):
    def __init__(self, name, extends=None, public=True, final=False, abstract=False):
        super(Interface, self).__init__(name, extends, public=public, final=final, abstract=False, interface=True)


class Enum(BaseClass):
    def __init__(self, name, extends=None, public=True, final=False):
        super(Enum, self).__init__(name, extends, public=public, final=final, enum=True)



from unittest import expectedFailure

from ..utils import TranspileTestCase


class GeneratorTests(TranspileTestCase):
    def test_simple_generator(self):
        self.assertCodeExecution("""
            def multiplier(first, second):
                y = first * second
                yield y
                y *= second
                yield y
                y *= second
                yield y
                y *= second
                yield y

            print(list(multiplier(1, 20)))
            """)

    def test_loop_generator(self):
        self.assertCodeExecution("""
            def fizz_buzz(start, stop):
                for i in range(start, stop):
                    found = False
                    if i % 2 == 0:
                        yield 'fizz'
                        found = True
                    if i % 3 == 0:
                        yield 'buzz'
                        found = True
                    if not found:
                        yield i

            print(list(fizz_buzz(1, 20)))
            """)

    def test_simplest_generator(self):
        self.assertCodeExecution("""
            def somegen():
                yield 1
                yield 2
                yield 3

            for i in somegen():
                print(i)
            """)

    def test_generator_method(self):
        self.assertCodeExecution("""
            class Interview:
                def __init__(self, start, stop):
                    self.start = start
                    self.stop = stop

                def fizz_buzz(self):
                    for i in range(self.start, self.stop):
                        found = False
                        if i % 2 == 0:
                            yield 'fizz'
                            found = True
                        if i % 3 == 0:
                            yield 'buzz'
                            found = True
                        if not found:
                            yield i

            myinterview = Interview(1, 20)

            for i in myinterview.fizz_buzz():
                print(i)
            """)

    def test_simple_generators(self):
        self.assertCodeExecution("""
            num_fibs = [5, 10, 15]

            def fib(n):
                a, b = 0, 1
                for _ in range(n):
                    yield a
                    a, b = b, a + b

            print([list(fib(n)) for n in num_fibs])
        """)

    def test_generator_send(self):
        self.assertCodeExecution("""
            def gen():
                a = yield
                print(a)

            g = gen()
            next(g)
            try:
                g.send("Hello World")
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_multi_send(self):
        self.assertCodeExecution("""
            def gen():
                a = yield 1
                print(a)
                b = yield 2
                print(b)

            g = gen()
            next(g)
            try:
                print(g.send("a"))
                print(g.send("b"))
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_send_loop(self):
        self.assertCodeExecution("""
            def gen():
                for i in range(1, 5):
                    a = yield i
                    print("printing from generator " + str(a))

            g = gen()
            g.send(None)
            try:
                while True:
                    b = g.send(1)
                    print("printing from user " + str(b))
                    print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_send_after_yield_stmt(self):
        self.assertCodeExecution("""
            def gen():
                yield "regular yield statement"
                a = yield "yield expression"
                print(a)

            g = gen()
            print(next(g))
            print(g.send("Hello World"))
            try:
                print(next(g))  # a is None
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_send_non_None(self):
        self.assertCodeExecution("""
            def gen():
                a = yield
                print(a)

            g = gen()
            try:
                g.send(1)
                print("should not print this")
            except Exception as e:
                print(e, e.args)
            """)

    def test_generator_yield_expr_call(self):
        self.assertCodeExecution("""
            def gen():
                while True:
                    x = print((yield))

            g = gen()
            g.send(None)
            g.send("Hello World")
            """)

    def test_generator_yield_expr_unary_op(self):
        self.assertCodeExecution("""
            def gen():
                while True:
                    x = -(yield)
                    yield x

            g = gen()
            g.send(None)
            g.send(2)
            g.send(-1)
            """)

    def test_generator_yield_expr_bool_op(self):
        self.assertCodeExecution("""
            def gen():
                while True:
                    x = not (yield)
                    yield x

            g = gen()
            g.send(None)
            g.send(True)
            g.send(False)
            """)

    def test_generator_yield_expr_binary_op(self):
        self.assertCodeExecution("""
            def gen():
                while True:
                    x = 2 + (yield)
                    yield x

            def gen2():
                while True:
                    x = (yield) * 2
                    yield x

            def gen3():
                while True:
                    x = (yield) ** 2
                    yield x

            g = gen()
            g.send(None)
            g.send(2)
            g.send(-1)

            g = gen2()
            g.send(None)
            g.send(100)
            g.send(20)

            g = gen3()
            next(g)
            g.send(2)
            g.send(0)
            """)

    def test_generator_yield_expr_aug_assign(self):
        self.assertCodeExecution("""
            def gen():
                while True:
                    x = 2
                    x += (yield)
                    yield x

            g = gen()
            g.send(None)
            g.send(1)
            g.send(2)
            """)

    def test_generator_yield_expr_compare(self):
        self.assertCodeExecution("""
            def gen():
                while True:
                    x = 1 <= (yield) < 5
                    yield x

            g = gen()
            g.send(None)
            g.send(1)
            g.send(100)
            """)

    def test_generator_yield_expr_if(self):
        self.assertCodeExecution("""
            def gen():
                if (yield):
                    print("Hello world")

            g = gen()
            g.send(None)
            try:
                g.send(1)
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_yield_expr_return(self):
        self.assertCodeExecution("""
            def gen():
                return (yield)

            g = gen()
            g.send(None)
            try:
                g.send(1)
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    @expectedFailure
    def test_generator_yield_try_finally_special_case(self):
        """Output is 'finally' followed by 'Hello World'
        due to the way CPython handles garbage collection."""
        self.assertCodeExecution("""
            def gen():
                try:
                    yield 'Hello World'
                finally:
                    print('finally')

            print(next(gen()))
            """)

    def test_generator_throw_on_starting(self):
        self.assertCodeExecution("""
            def gen():
                yield "Hello World"

            g = gen()
            try:
                g.throw(ZeroDivisionError)
                print("should not print this")
            except ZeroDivisionError:
                print("ZeroDivisionError")
            """)

    def test_generator_throw_other_exception(self):
        self.assertCodeExecution("""
            def gen():
                try:
                    yield
                    print("Hello World")
                except ZeroDivisionError:
                    raise TypeError

            g = gen()
            next(g)
            try:
                g.throw(ZeroDivisionError)
                print("should not print this")
            except TypeError:
                print("TypeError")
            """)

    def test_generator_throw_exception_handling(self):
        self.assertCodeExecution("""
            def gen():
                try:
                    yield "from try block"
                except TypeError:
                    yield "from catch block"

                yield "from outside try-except"
                a = yield

            g = gen()
            print(next(g))
            print(g.throw(TypeError))
            print(next(g))
            print(g.send("message"))
            """)

    def test_generator_throw_on_close(self):
        self.assertCodeExecution("""
            def gen():
                yield "Hello World"

            g = gen()
            g.close()
            try:
                g.throw(ZeroDivisionError)
                print("should not print this")
            except ZeroDivisionError:
                print("ZeroDivisionError")
            """)

    def test_generator_next_after_throw(self):
        self.assertCodeExecution("""
            def gen():
                yield 1
                yield 2

            g = gen()
            print(next(g))
            try:
                g.throw(TypeError)
                print("should not print this")
            except TypeError:
                try:
                    print(next(g))
                    print("should not print this")
                except StopIteration:
                    print("StopIteration")
            """)

    def test_generator_throw_args(self):
        self.assertCodeExecution("""
            def gen():
                yield 1

            g = gen()
            try:
                g.throw(TypeError, "Hello World")
                print("should not print this")
            except TypeError as e:
                print(e.__str__())

            g = gen()
            try:
                g.throw(ZeroDivisionError, 100)
                print("should not print this")
            except ZeroDivisionError as e:
                print(e.__str__())

            g = gen()
            try:
                g.throw(TypeError, (1, 2, 3, "Hello", "World"))
                print("should not print this")
            except TypeError as e:
                print(e.args)
            """)

    @expectedFailure
    def test_generator_throw_kwargs(self):
        self.assertCodeExecution("""
            def gen():
                yield 1
            g = gen()
            try:
                g.throw(TypeError, {"Hello": 1, "World": 2})
                print("should not print this")
            except TypeError as e:
                print(e.args)
            """)

    def test_generator_close(self):
        self.assertCodeExecution("""
            def gen():
                print("Hello world")
                try:
                    yield
                except TypeError:
                    pass
                except ZeroDivisionError:
                    pass

            g = gen()
            print(g.close())
            try:
                print(next(g))
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_close_ignore_exit(self):
        self.assertCodeExecution("""
            def gen():
                try:
                    yield
                except TypeError:
                    pass
                except GeneratorExit:
                    yield "exit ignored"

            g = gen()
            next(g)
            try:
                g.close()
                print("should not print this")
            except RuntimeError:
                print(RuntimeError)
            """)

    def test_generator_close_exception_propagation(self):
        self.assertCodeExecution("""
            def gen():
                try:
                    yield
                except GeneratorExit:
                    raise OSError

            g = gen()
            next(g)
            try:
                g.close()
                print("should not print this")
            except OSError:
                pass
            """)

    def test_generator_close_twice(self):
        self.assertCodeExecution("""
            def gen():
                yield 1
                yield 2

            g = gen()
            print(g.close())
            print(g.close())
            """)

    def test_generator_yield_try_suite(self):
        self.assertCodeExecution("""
            def gen():
                try:
                    yield "from try block"
                except:
                    yield "from except block"
                else:
                    yield "from else block"
                finally:
                    yield "from finally block"

            g = gen()
            print(next(g))
            print(g.throw(TypeError))
            print(next(g))

            g = gen()
            print(next(g))
            print(next(g))
            print(next(g))
            try:
                print(next(g))
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_yield_expr_try_suite(self):
        self.assertCodeExecution("""
            def gen():
                try:
                    yield "from try block"
                    a = yield
                except:
                    yield "from except block"
                    a = yield
                else:
                    a = yield "from else block"
                finally:
                    a = yield "from finally block"
                print(a)

            g = gen()
            print(next(g))
            print(g.throw(TypeError))
            print(next(g))
            print(g.send("except"))
            try:
                print(g.send("finally"))
                print("should not print this")
            except StopIteration:
                print("StopIteration")

            g = gen()
            print(next(g))
            print(next(g))
            print(g.send("try"))
            try:
                print(g.send("finally"))
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_simplest_yieldfrom(self):
        self.assertCodeExecution("""
            def gen1():
                yield 1
                yield 2
                yield 3

            def gen2():
                yield from gen1()

            for i in gen2():
                print(i)
        """)

    def test_yieldfrom_list(self):
        self.assertCodeExecution("""
            def gen():
                yield from [1, 2, 3]

            for i in gen():
                print(i)
        """)

    def test_chaining_yieldfrom(self):
        self.assertCodeExecution("""
            def gen1():
                yield 'a'
                yield 'b'

            def gen2():
                yield from gen1()
                yield 'gen2'

            def gen3():
                yield from gen2()

            for i in gen3():
                print(i)
        """)

    def test_generator_yieldfrom_generator_exit(self):
        self.assertCodeExecution("""
            def gen1():
                return [1, 2, 3]

            def gen2():
                v = yield from gen1()
                yield 1

            g = gen2()
            print(next(g))
            try:
                print(g.throw(GeneratorExit))
                print("should not print this")
            except GeneratorExit:
                try:
                    print(next(g))
                    print("should not print this")
                except StopIteration:
                    print("StopIteration")
            """)

    def test_generator_yieldfrom_throw_propagation(self):
        self.assertCodeExecution("""
            def gen1():
                try:
                    yield 1
                except TypeError:
                    yield "TypeError"

            def gen2():
                yield from gen1()

            g = gen2()
            print(next(g))
            print(g.throw(TypeError))
            """)

    def test_generator_yieldfrom_send_propagation(self):
        self.assertCodeExecution("""
            def gen1():
                a = yield "gen1"
                print("value received: " + a)

            def gen2():
                yield from gen1()

            g = gen2()
            print(next(g))
            try:
                print(g.send("Hello World"))
                print("should not print this")
            except StopIteration:
                print("StopIteration")
            """)

    def test_generator_yieldfrom_return(self):
        self.assertCodeExecution("""
            def gen():
                return (yield from [1, 2, 3])
                return "impossible return"

            g = gen()
            for i in g:
                print(i)
            try:
                print(next(g))
                print("should not print this")
            except StopIteration as e:
                print(e)
            """)

    def test_generator_yieldfrom_in_loop(self):
        self.assertCodeExecution("""
            def gen():
                a = [1, 2, 3]
                while len(a):
                    yield from a
                    a = []

            g = gen()
            for i in g:
                print(i)
            """)

    def test_generator_stop_iteration_value(self):
        self.assertCodeExecution("""
            def gen():
                yield "Hello World"
                return "return value"

            g = gen()
            print(next(g))
            try:
                print(next(g))
                print("should not print this")
            except StopIteration as e:
                print(e.value)

            def gen1():
                yield "Hello World"
                return "return value"

            def gen2():
                v = yield from gen1()
                print(v)

            g = gen2()
            print(next(g))
            try:
                print(next(g))
                print("should not print this")
            except StopIteration:
                print("StopIteration")

            def gen3():
                yield 1
                yield 2
                return 10

            g1 = gen3()
            print(next(g1))
            print(next(g1))
            try:
                print(next(g1))
                print("should not print this")
            except StopIteration as e:
                print(e.value)

            g2 = gen3()
            for i in g2:
                print(i)
            try:
                print(next(g2))
                print("should not print this")
            except StopIteration as e:
                print(e.value)
            """)



from ..utils import TranspileTestCase


class DecoratorTests(TranspileTestCase):
    def test_simple_decorator(self):
        self.assertCodeExecution("""
            def thing(fn):
                def _dec(value):
                    print("Start decoration.")
                    result = fn(value)
                    print("End decoration.")
                    return result * 42
                return _dec

            @thing
            def calculate(value):
                print("Do a calculation")
                return 37 * value

            print("Decorated result is", calculate(5))
            print("Done.")
            """)

    def test_decorator_with_argument(self):
        self.assertCodeExecution("""
            def thing(multiplier):
                def _dec(fn):
                    def _fn(value):
                        print("Start decoration.")
                        result = fn(value)
                        print("End decoration.")
                        return result * multiplier
                    return _fn
                return _dec

            @thing(42)
            def calculate(value):
                print("Do a calculation")
                return 37 * value

            print("Decorated result is", calculate(5))
            print("Done.")
            """)



from unittest import expectedFailure

from ..utils import TranspileTestCase


class AssignmentTests(TranspileTestCase):
    def test_simple_assignment(self):
        self.assertCodeExecution("""
            x = 42
            print(x)
            """)

    def test_multiple_assignment(self):
        self.assertCodeExecution("""
            x = y = 42
            print(x, y)
            """)

    def test_old_style_conditional_assignment(self):
        self.assertCodeExecution("""
            x = 42
            y = x or 37
            print(y)
            x = 0
            y = x or 37
            print(y)
            """)

    def test_conditional_assignment(self):
        self.assertCodeExecution("""
            x = 42
            y = 99 if x else 37
            print(y)
            x = 0
            y = 99 if x else 37
            print(y)
            """)

    def test_access_potentially_unassigned(self):
        self.assertCodeExecution("""
            x = 37
            if x > 0:
                y = 42
            print(y)
            """)

    def test_use_potentially_unassigned(self):
        self.assertCodeExecution("""
            x = 37
            try:
                if y > 0:
                    print("Yes")
                else:
                    print("No")
            except NameError as err:
                print(err)
            """)

    def test_assign_to_argument(self):
        self.assertCodeExecution("""
            def foo(arg):
                val = arg + 10
                arg = val - 2
                return arg
            print(foo(20))
            """)

    def test_list_assignment(self):
        self.assertCodeExecution("""
            [x, y, z] = range(3)
            print(x)
            print(y)
            print(z)
        """)

    @expectedFailure
    def test_bad_list_assignment(self):
        self.assertCodeExecution("""
            try:
                [x, y, z, a] = range(3)
                print(x)
                print(y)
                print(z)
                print(a)
            except ValueError as e:
                print("Got a ValueError:", e)
        """)

    def test_bad_list_assignment_raises_StopIteration(self):
        """
        This test is a copy of test_bad_list_assignment.
        Raising ValueError as a result of a bad unpacking is not currently implemented;
        this test should be deleted when it is.
        """
        self.assertCodeExecution("""
            try:
                [x, y, z, a] = range(3)
                print(x)
                print(y)
                print(z)
                print(a)
            except (StopIteration, ValueError) as e:
                print("ValueError: not enough values to unpack (expected 4, got 3)")
        """)

    def test_tuple_assignment(self):
        self.assertCodeExecution("""
            (x, y, z) = range(3)
            print(x)
            print(y)
            print(z)
            """)

    @expectedFailure
    def test_bad_tuple_assignment(self):
        self.assertCodeExecution("""
            try:
                (x, y, z, a) = range(3)
                print(x)
                print(y)
                print(z)
                print(a)
            except ValueError as e:
                print("Got a ValueError:", e)
        """)

    def test_bad_tuple_assignment_raises_StopIteration(self):
        """
        This test is a copy of test_bad_tuple_assignment.
        Raising ValueError as a result of a bad unpacking is not currently implemented;
        this test should be deleted when it is.
        """
        self.assertCodeExecution("""
            try:
                (x, y, z, a) = range(3)
                print(x)
                print(y)
                print(z)
                print(a)
            except (StopIteration, ValueError) as e:
                print("ValueError: not enough values to unpack (expected 4, got 3)")
        """)

    def test_implied_tuple_assignment(self):
        self.assertCodeExecution("""
            x, y, z = range(3)
            print(x)
            print(y)
            print(z)
            """)

    @expectedFailure
    def test_bad_implied_tuple_assignment(self):
        self.assertCodeExecution("""
            try:
                x, y, z, a = range(3)
                print(x)
                print(y)
                print(z)
                print(a)
            except ValueError as e:
                print("Got a ValueError:", e)
        """)

    def test_bad_implied_tuple_assignment_raises_StopIteration(self):
        """
        This test is a copy of test_bad_implied_tuple_assignment.
        Raising ValueError as a result of a bad unpacking is not currently implemented;
        this test should be deleted when it is.
        """
        self.assertCodeExecution("""
            try:
                x, y, z, a = range(3)
                print(x)
                print(y)
                print(z)
                print(a)
            except (StopIteration, ValueError) as e:
                print("ValueError: not enough values to unpack (expected 4, got 3)")
        """)

    def test_increment_assignment(self):
        self.assertCodeExecution("""
            a = 1
            a += 1
            print(a)
            """)

    def test_increment_assignment_attribute(self):
        self.assertCodeExecution("""
            class Thing(object):
                a = 1

            t = Thing()
            t.a += 1
            print(t.a)
            """)

    def test_increment_assignment_subscript(self):
        self.assertCodeExecution("""
            a = [0, 1, 2]
            a[0] += 1
            print(a)
            """)



import ast

from ..java import opcodes as JavaOpcodes
from .types.primitives import ALOAD_name

##########################################################################
# Debugging helpers
##########################################################################


def dump(node, annotate_fields=True, include_attributes=True, indent='  '):
    """
    Return a formatted dump of the tree in *node*.  This is mainly useful for
    debugging purposes.  The returned string will show the names and the values
    for fields.  This makes the code impossible to evaluate, so if evaluation is
    wanted *annotate_fields* must be set to False.  Attributes such as line
    numbers and column offsets are not dumped by default.  If this is wanted,
    *include_attributes* can be set to True.
    """
    def _format(node, level=0):
        if isinstance(node, ast.AST):
            fields = [(a, _format(b, level)) for a, b in ast.iter_fields(node)]
            if include_attributes and node._attributes:
                fields.extend([(a, _format(getattr(node, a), level))
                               for a in node._attributes])
            return ''.join([
                node.__class__.__name__,
                '(',
                ', '.join(
                    ('%s=%s' % field for field in fields)
                    if annotate_fields else (b for a, b in fields)),
                ')'])
        elif isinstance(node, list):
            lines = ['[']
            lines.extend((indent * (level + 2) + _format(x, level + 2) + ','
                         for x in node))
            if len(lines) > 1:
                lines.append(indent * (level + 1) + ']')
            else:
                lines[-1] += ']'
            return '\n'.join(lines)
        return repr(node)

    if not isinstance(node, ast.AST):
        raise TypeError('expected AST, got %r' % node.__class__.__name__)

    return _format(node)


class DEBUG:
    def __init__(self, msg):
        self.msg = msg

    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.LDC_W(self.msg),
            JavaOpcodes.INVOKESTATIC(
                'org/Python',
                'debug',
                args=['Ljava/lang/String;'],
                returns='V'),
        )
        # This opcode isn't for the final output.
        return False


class DEBUG_name:
    def __init__(self, name):
        self.name = name

    def process(self, context):
        context.add_opcodes(
            JavaOpcodes.LDC_W(self.name),
            ALOAD_name(self.name),
            JavaOpcodes.INVOKESTATIC(
                'org/Python',
                'debug',
                args=['Ljava/lang/String;', 'Ljava/lang/Object;'],
                returns='V'
            ),
        )
        # This opcode isn't for the final output.
        return False


class DEBUG_value:
    def __init__(self, msg, dup=False):
        self.msg = msg
        self.dup = dup

    def process(self, context):
        if self.dup:
            context.add_opcodes(JavaOpcodes.DUP())
        context.add_opcodes(
            JavaOpcodes.LDC_W(self.msg),
            JavaOpcodes.SWAP(),
            JavaOpcodes.INVOKESTATIC(
                'org/Python',
                'debug',
                args=['Ljava/lang/String;', 'Ljava/lang/Object;'],
                returns='V'
            ),
        )
        # This opcode isn't for the final output.
        return False



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class PropertyTests(TranspileTestCase):
    pass


class BuiltinPropertyFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["property"]

    not_implemented = [
    ]



from unittest import expectedFailure

from ..utils import TranspileTestCase


class ClassTests(TranspileTestCase):
    def test_minimal(self):
        self.assertCodeExecution("""
            class MyClass:
                pass

            obj = MyClass()
            """, run_in_function=False)

    def test_simple(self):
        self.assertCodeExecution("""
            class MyClass:
                def __init__(self, val):
                    print("VAL: ", val)
                    self.value = val

                def stuff(self, delta):
                    print("DELTA: ", delta)
                    return self.value + delta

            obj = MyClass(4)
            obj.stuff(5)
            """, run_in_function=False)

    def test_subclass_object(self):
        self.assertCodeExecution("""
            class MyClass(object):
                def __init__(self, val):
                    print("VAL: ", val)
                    self.value = val

                def stuff(self, delta):
                    print("DELTA: ", delta)
                    return self.value + delta

            obj = MyClass(4)
            obj.stuff(5)
            """, run_in_function=False)

    def test_method_override(self):
        self.assertCodeExecution("""
            class MyObject:
                def __init__(self, x):
                    self.x = x

                def __str__(self):
                    return "Myobject instance %s" % self.x

            obj = MyObject(37)

            print(obj)
            """, run_in_function=False)

    def test_subclass(self):
        self.assertCodeExecution("""
            class MyBase:
                def __init__(self, x):
                    self.x = x

                def __str__(self):
                    return "Mybase instance %s" % self.x

                def first(self):
                    return self.x * 2


            class MyObject(MyBase):
                def __init__(self, x, y):
                    super().__init__(x)
                    self.y = y

                def __str__(self):
                    return "Myobject instance %s, %s" % (self.x, self.y)

                def second(self):
                    return self.x * self.y

            obj = MyObject(37, 42)

            print(obj)
            print(obj.x)
            print(obj.first())
            print(obj.y)
            print(obj.second())
            """)

    def test_subclass_2_clause_super(self):
        self.assertCodeExecution("""
            class MyBase:
                def __init__(self, x):
                    self.x = x

                def __str__(self):
                    return "Mybase instance %s" % self.x

                def first(self):
                    return self.x * 2


            class MyObject(MyBase):
                def __init__(self, x, y):
                    super(MyObject, self).__init__(x)
                    self.y = y

                def __str__(self):
                    return "Myobject instance %s, %s" % (self.x, self.y)

                def second(self):
                    return self.x * self.y

            obj = MyObject(37, 42)

            print(obj)
            print(obj.x)
            print(obj.first())
            print(obj.y)
            print(obj.second())
            """)

    def test_redefine(self):
        self.assertCodeExecution("""
            class MyClass:
                def __init__(self, val):
                    print("VAL: ", val)
                    self.value = val

                def stuff(self, delta):
                    print("DELTA: ", delta)
                    return self.value + delta

                def stuff(self, delta):
                    print("Redefined DELTA: ", delta)
                    return self.value + delta * 2

            obj = MyClass(4)
            obj.stuff(5)
            """, run_in_function=False)

    def test_overwrite_class_attributes(self):
        self.assertCodeExecution("""
            class MyClass:
                x = 1

            print(MyClass.x)

            inst = MyClass()
            print(inst.x)

            inst.x = 123
            print(inst.x)

            print(MyClass.x)

            inst2 = MyClass()
            inst2.x = 5
            print(inst2.x)
            """, run_in_function=False)

    def test_class_with_method_with_type_hints(self):
        self.assertCodeExecution("""
            class MyClass():
                def foo(self, index: int) -> int:
                    return index
        """)


class ClassMethodTests(TranspileTestCase):
    @expectedFailure
    def test_classmethod(self):
        self.assertCodeExecution("""
            class MyClass:
                @classmethod
                def foo(cls, arg1, arg2):
                    print("This is a classmethod on ", cls, arg1, arg2)

            obj = MyClass()
            obj.foo(1, 2)
            MyClass.foo(3, 4)
            """, run_in_function=False)


class StaticMethodTests(TranspileTestCase):
    @expectedFailure
    def test_staticmethod(self):
        self.assertCodeExecution("""
            class MyClass:
                @staticmethod
                def foo(arg1, arg2):
                    print("This is a staticmethod ", arg1, arg2)

            obj = MyClass()
            obj.foo(1, 2)
            MyClass.foo(3, 4)
            """, run_in_function=False)


class InnerClassTests(TranspileTestCase):
    def test_inner_simple(self):
        self.assertCodeExecution("""
            class MyClass:
                class InnerClass:
                    def __init__(self, val):
                        print("VAL: ", val)
                        self.value = val

                    def stuff(self, delta):
                        print("DELTA: ", delta)
                        return self.value + delta

            obj = MyClass.InnerClass(4)
            obj.stuff(5)
        """, run_in_function=False)

    def test_inner_namespaced(self):
        self.assertCodeExecution("""
            class MyClass1:
                class InnerClass:
                    def stuff(self):
                        print('STUFF FROM 1')

            class MyClass2:
                class InnerClass:
                    def stuff(self):
                        print('STUFF FROM 2')

            obj = MyClass1.InnerClass()
            obj.stuff()
            obj = MyClass2.InnerClass()
            obj.stuff()
        """, run_in_function=False)



#!/usr/bin/env python

"""Microbenchmark for Python's sequence unpacking."""

import time


def do_unpacking(iterations, to_unpack):
    times = []
    for _ in range(iterations):
        t0 = time.time()
        # Should be 400 unpackings, but MethodCodeTooLarge
        # TODO Look into why this code is so big
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        """
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        a, b, c, d, e, f, g, h, i, j = to_unpack
        """
        t1 = time.time()
        times.append(t1 - t0)
    return times


def test_tuple_unpacking(iterations):
    x = tuple(range(10))
    return do_unpacking(iterations, x)


def test_list_unpacking(iterations):
    x = range(10)

    return do_unpacking(iterations, x)


def test_all(iterations):
    tuple_data = test_tuple_unpacking(iterations)
    list_data = test_list_unpacking(iterations)

    return [x + y for (x, y) in zip(tuple_data, list_data)]


if __name__ == "__main__":
    import sys
    loops = int(sys.argv[1])
    times = test_all(loops)

    print("Time elapsed: " + str(sum(times)) + " sec")



import atexit
import contextlib
from io import StringIO, BytesIO
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time
import traceback
from unittest import TestCase
from itertools import permutations

from voc.python.blocks import Block as PyBlock
from voc.python.modules import Module as PyModule
from voc.java.constants import ConstantPool, Utf8
from voc.java.klass import ClassFileReader, ClassFileWriter

from voc.java.attributes import Code as JavaCode
from voc.transpiler import Transpiler

# get path to `tests` directory
TESTS_DIR = os.path.dirname(__file__)

# A state variable to determine if the test environment has been configured.
_suite_configured = False

# Temporary directory containing all files generated by this test process
# Set during setUpSuite()
_output_dir = ''


def setUpSuite():
    """Configure the entire test suite.

    This only needs to be run once, prior to the first test.
    """
    global _output_dir
    global _suite_configured
    if _suite_configured:
        return

    def remove_output_dir():
        global _output_dir
        if _output_dir != '':
            try:
                shutil.rmtree(_output_dir)
            except FileNotFoundError:
                pass

    atexit.register(remove_output_dir)
    _output_dir = tempfile.mkdtemp(dir=TESTS_DIR)

    os.environ['VOC_BUILD_DIR'] = os.path.join(_output_dir, 'build')
    os.environ['VOC_DIST_DIR'] = os.path.join(_output_dir, 'dist')
    proc = subprocess.Popen(
        "ant java",
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
    )

    try:
        out, err = proc.communicate(timeout=30)
    except subprocess.TimeoutExpired:
        proc.kill()
        out, err = proc.communicate()
        raise

    if proc.returncode != 0:
        raise Exception("Error compiling java sources: " + out.decode('ascii'))

    _suite_configured = True


@contextlib.contextmanager
def capture_output(redirect_stderr=True):
    oldout, olderr = sys.stdout, sys.stderr
    try:
        out = StringIO()
        sys.stdout = out
        if redirect_stderr:
            sys.stderr = out
        else:
            sys.stderr = StringIO()
        yield out
    except:
        if redirect_stderr:
            traceback.print_exc()
        else:
            raise
    finally:
        sys.stdout, sys.stderr = oldout, olderr


def adjust(text, run_in_function=False):
    """Adjust a code sample to remove leading whitespace."""
    lines = text.split('\n')
    if len(lines) == 1:
        return text

    if lines[0].strip() == '':
        lines = lines[1:]
    first_line = lines[0].lstrip()
    n_spaces = len(lines[0]) - len(first_line)

    final_lines = [('    ' if run_in_function else '') + line[n_spaces:] for line in lines]

    if run_in_function:
        final_lines = [
            "def test_function():",
        ] + final_lines + [
            "test_function()",
        ]

    return '\n'.join(final_lines)


def runAsPython(test_dir, main_code, extra_code=None, args=None):
    """Run a block of Python code with the Python interpreter."""
    # Output source code into test directory
    with open(os.path.join(test_dir, 'test.py'), 'w', encoding='utf-8') as py_source:
        py_source.write(main_code)

    if extra_code:
        for name, code in extra_code.items():
            path = name.split('.')
            path[-1] = path[-1] + '.py'
            if len(path) != 1:
                try:
                    os.makedirs(os.path.join(test_dir, *path[:-1]))
                except FileExistsError:
                    pass
            with open(os.path.join(test_dir, *path), 'w', encoding="utf-8") as py_source:
                py_source.write(adjust(code))

    if args is None:
        args = []

    proc = subprocess.Popen(
        [sys.executable, "test.py"] + args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=test_dir,
    )
    out = proc.communicate()

    return out[0].decode('utf8')


def compileJava(java_dir, java):
    if not java:
        return None

    sources = []
    with capture_output():
        for descriptor, code in java.items():
            parts = descriptor.split('/')

            class_dir = os.path.sep.join(parts[:-1])
            class_file = os.path.join(class_dir, "%s.java" % parts[-1])

            full_dir = os.path.join(java_dir, class_dir)
            full_path = os.path.join(java_dir, class_file)

            try:
                os.makedirs(full_dir)
            except FileExistsError:
                pass

            with open(full_path, 'w', encoding="utf-8") as java_source:
                java_source.write(adjust(code))

            sources.append(class_file)

    classpath = os.pathsep.join([
        os.path.join('..', 'dist', 'python-java-support.jar'),
        os.curdir,
    ])
    proc = subprocess.Popen(
        ["javac", "-classpath", classpath] + sources,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=java_dir,
    )
    out = proc.communicate()

    return out[0].decode('utf8')


JAVA_EXCEPTION = re.compile(
    '(((Exception in thread "[\w-]+" )?(?P<exception1>[\w]+): (?P<message1>[^\r?\n]+))|' +
    '([^\r\n]*?\r?\n((    |\t)at[^\r\n]*?\r?\n)*' +
    'Caused by: (?P<exception2>[\w]+)(?:\:\s)?(?P<message2>[^\r?]+))\r?)\n' +
    '(?P<trace>(\s+at .+\((((.*)(:(\d+))?)|(Native Method))\)\r?\n)+)(.*\r?\n)*' +
    '(Exception in thread "\w+" )?',
    re.MULTILINE
)
JAVA_STACK = re.compile(
    '^\s+at (?P<module>.+)\((((?P<file>.*?)(:(?P<line>\d+))?)|(Native Method))\)\r?\n',
    re.MULTILINE
)

PYDEV_CONNECT = re.compile("pydev debugger: process \\d+ is connecting[\r|\n]+")

PYTHON_EXCEPTION = re.compile(
    'Traceback \(most recent call last\):\r?\n(  File "(?P<file>.*)", line (?P<line>\d+), ' +
    'in .*\r?\n    .*\r?\n)+(?P<exception>[^:]*)(?::\s)?(?P<message>.*\n)$')

PYTHON_STACK = re.compile('  File "(?P<file>.*)", line (?P<line>\d+), in .*\r?\n    .*\r?\n')

MEMORY_REFERENCE = re.compile('0x[\dABCDEFabcdef]{4,16}')

END_OF_CODE_STRING = '===end of test==='
END_OF_CODE_STRING_NEWLINE = END_OF_CODE_STRING + '\n'

# Prevent floating point discrepancies in very low significant digits from being an issue
FLOAT_PRECISION = re.compile('(\\.\d{5})\d+')


def cleanse_java(raw, substitutions):
    matches = JAVA_EXCEPTION.search(raw)
    if matches is not None:
        groups = matches.groupdict()
        if groups['exception2'] is not None:
            # Test the specific message
            out = JAVA_EXCEPTION.sub(
                '### EXCEPTION ###{linesep}\\g<exception2>: \\g<message2>{linesep}\\g<trace>'.format(
                    linesep=os.linesep
                ),
                raw
            )
        else:
            # Test the specific message
            out = JAVA_EXCEPTION.sub(
                '### EXCEPTION ###{linesep}\\g<exception1>: \\g<message1>{linesep}\\g<trace>'.format(
                    linesep=os.linesep
                ),
                raw
            )
    else:
        out = raw

    stack = JAVA_STACK.findall(out)
    out = JAVA_STACK.sub('', out)

    out = '%s%s%s' % (
        out,
        os.linesep.join([
            "    %s:%s" % (s[3], s[5])
            for s in stack[::-1]
            if s[0].startswith('python.') and not s[0].endswith('.<init>') and s[5]
        ]),
        os.linesep if stack else ''
    )

    out = MEMORY_REFERENCE.sub("0xXXXXXXXX", out)

    out = out.replace(
        "'python.test'", '***EXECUTABLE***').replace(
        "'python.testdaemon.TestDaemon'", '***EXECUTABLE***')

    # Replace references to the test script with something generic
    out = out.replace("'test.py'", '***EXECUTABLE***')

    # Replace all the explicit data substitutions
    if substitutions:
        for to_value, from_values in substitutions.items():
            for from_value in from_values:
                out = out.replace(from_value, to_value)

    out = out.replace('\r\n', '\n')

    # Replace high precision floats with abbreviated forms
    out = FLOAT_PRECISION.sub('\\1...', out)

    return out


def cleanse_python(raw, substitutions):
    # Test the specific message
    out = PYTHON_EXCEPTION.sub(
        '### EXCEPTION ###{linesep}\\g<exception>: \\g<message>'.format(linesep=os.linesep),
        raw
    )

    out = PYDEV_CONNECT.sub('', out)

    stack = PYTHON_STACK.findall(raw)
    out = '%s%s%s' % (
        out,
        os.linesep.join(
            [
                "    %s:%s" % (s[0], s[1])
                for s in stack
            ]
        ),
        os.linesep if stack else ''
    )
    # Normalize memory references from output
    out = MEMORY_REFERENCE.sub("0xXXXXXXXX", out)

    # Replace references to the test script with something generic
    out = out.replace("'test.py'", '***EXECUTABLE***')

    if substitutions:
        for to_value, from_values in substitutions.items():
            for from_value in from_values:
                out = out.replace(from_value, to_value)

    out = out.replace('\r\n', '\n')

    # Replace high precision floats with abbreviated forms
    out = FLOAT_PRECISION.sub('\\1...', out)

    return out


class TranspileTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        global _output_dir
        setUpSuite()
        cls.temp_dir = os.path.join(_output_dir, 'temp')
        classpath = os.pathsep.join([
            os.path.join('dist', 'python-java-testdaemon.jar'),
            os.path.join('dist', 'python-java-support.jar'),
        ])
        cls.jvm = subprocess.Popen(
            ["java", "-classpath", classpath, "python.testdaemon.TestDaemon"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            cwd=_output_dir,
        )

    @classmethod
    def tearDownClass(cls):
        if cls.jvm is not None:
            # use communicate here to wait for process to exit
            cls.jvm.communicate("exit".encode("utf-8"))

    def assertBlock(self, python, java):
        self.maxDiff = None
        dump = False

        py_block = PyBlock(parent=PyModule('test', 'test.py'))
        if python:
            python = adjust(python)
            code = compile(python, '<test>', 'exec')
            py_block.extract(code, debug=dump)

        java_code = py_block.transpile()

        out = BytesIO()
        constant_pool = ConstantPool()
        java_code.resolve(constant_pool)

        constant_pool.add(Utf8('test'))
        constant_pool.add(Utf8('Code'))
        constant_pool.add(Utf8('LineNumberTable'))

        writer = ClassFileWriter(out, constant_pool)
        java_code.write(writer)

        debug = StringIO()
        reader = ClassFileReader(BytesIO(out.getbuffer()), constant_pool, debug=debug)
        JavaCode.read(reader, dump=0)

        if dump:
            print(debug.getvalue())

        java = adjust(java)
        self.assertEqual(debug.getvalue(), java[1:])

    def assertCodeExecution(
            self, code,
            message=None,
            extra_code=None,
            run_in_global=True, run_in_function=True, exits_early=False,
            args=None, substitutions=None):
        """"Run code as native python, and under Java and check the output is identical"""
        self.maxDiff = None
        # ==================================================
        # Pass 1 - run the code in the global context
        # ==================================================
        if run_in_global:
            try:
                self.makeTempDir()
                # Run the code as Python and as Java.
                adj_code = adjust(code, run_in_function=False)
                adj_code += '\nprint("%s")\n' % END_OF_CODE_STRING
                py_out = runAsPython(self.temp_dir, adj_code, extra_code, args=args)
                java_out = self.runAsJava(adj_code, extra_code, args=args)
            except Exception as e:
                self.fail(e)
            finally:
                # Clean up the test directory where the class file was written.
                # (we have to set ignore_errors for this to work on Windows)
                # Ignoring all errors can be anxiety-inducing so we'll do one more
                # check on the directory afterwards now.
                shutil.rmtree(self.temp_dir, ignore_errors=True)

                if os.path.exists(self.temp_dir):
                    raise IsADirectoryError("{} was unsuccessfully deleted".format
                                            (self.temp_dir))
                # print(java_out)

            # Cleanse the Python and Java output, producing a simple
            # normalized format for exceptions, floats etc.
            java_out = cleanse_java(java_out, substitutions)
            py_out = cleanse_python(py_out, substitutions)

            # Confirm that the output of the Java code is the same as the Python code.
            if message:
                context = 'Global context: %s' % message
            else:
                context = 'Global context'
            self.assertEqual(java_out, py_out, context)

            # Confirm that both output strings end with the canary statement
            if exits_early:
                if java_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Java test failed to raise exception \n%s" % java_out)
                if py_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Python test failed to raise exception \n%s" % py_out)
            else:
                if not java_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Java test failed prematurely \n%s" % java_out)
                if not py_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Python test failed prematurely \n%s" % py_out)
        # ==================================================
        # Pass 2 - run the code in a function's context
        # ==================================================
        if run_in_function:
            try:
                self.makeTempDir()
                # Run the code as Python and as Java.
                adj_code = adjust(code, run_in_function=True)
                adj_code += '\nprint("%s")\n' % END_OF_CODE_STRING
                py_out = runAsPython(self.temp_dir, adj_code, extra_code, args=args)
                java_out = self.runAsJava(adj_code, extra_code, args=args)
            except Exception as e:
                self.fail(e)
            finally:
                # Clean up the test directory where the class file was written.
                shutil.rmtree(self.temp_dir)
                # print(java_out)

            # Cleanse the Python and Java output, producing a simple
            # normalized format for exceptions, floats etc.
            java_out = cleanse_java(java_out, substitutions)
            py_out = cleanse_python(py_out, substitutions)

            # Confirm that the output of the Java code is the same as the Python code.
            if message:
                context = 'Function context: %s' % message
            else:
                context = 'Function context'
            self.assertEqual(java_out, py_out, context)

            # Confirm that both output strings end with the canary statement
            if exits_early:
                if java_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Java test failed to raise exception \n%s" % java_out)
                if py_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Python test failed to raise exception \n%s" % py_out)
            else:
                if not java_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Java test failed prematurely \n%s" % java_out)
                if not py_out.endswith(END_OF_CODE_STRING_NEWLINE):
                    self.fail("Python test failed prematurely \n%s" % py_out)

    def assertJavaExecution(
                self, code, out,
                extra_code=None, java=None,
                run_in_global=True, run_in_function=True,
                args=None, substitutions=None):
        """Run code under Java and check the output is as expected"""
        global _output_dir
        self.maxDiff = None
        try:
            # ==================================================
            # Prep - compile any required Java sources
            # ==================================================
            # Create the temp directory into which code will be placed
            java_dir = os.path.join(_output_dir, 'java')

            try:
                os.makedirs(java_dir)
            except FileExistsError:
                pass

            # Compile the java support code
            java_compile_out = compileJava(java_dir, java)

            if java_compile_out:
                self.fail(java_compile_out)

            # Cleanse the Python output, producing a simple
            # normalized format for exceptions, floats etc.
            py_out = adjust(out)

            # ==================================================
            # Pass 1 - run the code in the global context
            # ==================================================
            if run_in_global:
                try:
                    self.makeTempDir()
                    # Run the code as Java.
                    java_out = self.runAsJava(adjust(code), extra_code, args=args)
                except Exception as e:
                    self.fail(e)
                finally:
                    # Clean up the test directory where the class file was written.
                    shutil.rmtree(self.temp_dir)
                    # print(java_out)

                # Cleanse the Java output, producing a simple
                # normalized format for exceptions, floats etc.
                java_out = cleanse_java(java_out, substitutions)

                # Confirm that the output of the Java code is the same as the Python code.
                self.assertEqual(java_out, py_out, 'Global context')

            # ==================================================
            # Pass 2 - run the code in a function's context
            # ==================================================
            if run_in_function:
                try:
                    self.makeTempDir()
                    # Run the code as Java.
                    java_out = self.runAsJava(
                        adjust(code, run_in_function=True),
                        extra_code, args=args)
                except Exception as e:
                    self.fail(e)
                finally:
                    # Clean up the test directory where the class file was written.
                    shutil.rmtree(self.temp_dir)
                    # print(java_out)

                # Cleanse the Java output, producing a simple
                # normalized format for exceptions, floats etc.
                java_out = cleanse_java(java_out, substitutions)

                # Confirm that the output of the Java code is the same as the Python code.
                self.assertEqual(java_out, py_out, 'Function context')

        finally:
            # Clean up the java directory where the class file was written.
            if os.path.exists(java_dir):
                shutil.rmtree(java_dir)

    def makeTempDir(self):
        """Create a "temp" subdirectory in the class's generated temporary directory if it doesn't currently exist."""
        try:
            os.mkdir(self.temp_dir)
        except FileExistsError:
            pass

    def runAsJava(self, main_code, extra_code=None, args=None, timed=False):
        """Run a block of Python code as a Java program."""
        # Output source code into test directory
        transpiler = Transpiler(verbosity=0)

        # Don't redirect stderr; we want to see any errors from the transpiler
        # as top level test failures.
        with capture_output(redirect_stderr=False):

            transpiler.transpile_string("test.py", main_code)

            if extra_code:
                for name, code in extra_code.items():
                    transpiler.transpile_string("%s.py" % name.replace('.', os.path.sep), adjust(code))

        transpiler.write(self.temp_dir)

        if args is None:
            args = []

        t1_start = time.perf_counter()
        t2_start = time.process_time()

        if len(args) == 0:

            # encode to turn str into bytes-like object
            self.jvm.stdin.write(("python.test\n").encode("utf-8"))
            self.jvm.stdin.flush()

            out = ""
            while True:
                try:
                    line = self.jvm.stdout.readline().decode("utf-8")
                    if line == ".{0}".format(os.linesep):
                        break
                    else:
                        out += line
                except IOError as e:
                    continue

            t1_stop = time.perf_counter()
            t2_stop = time.process_time()

        else:
            classpath = os.pathsep.join([
                os.path.join('..', 'dist', 'python-java-support.jar'),
                os.path.join('..', 'java'),
                os.curdir,
            ])

            proc = subprocess.Popen(
                ["java", "-classpath", classpath, "python.test"] + args,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                cwd=self.temp_dir
            )

            t1_stop = time.perf_counter()
            t2_stop = time.process_time()

            out = proc.communicate()[0].decode('utf8')

            if proc.returncode != 0:
                raise Exception(
                    "Java subprocess didn't exit cleanly (exit status %s)\n\n: %s" % (
                        proc.returncode, out
                    )
                )

        if timed:
            print("  Elapsed time: ", (t1_stop-t1_start), " sec")
            print("  CPU process time: ", (t2_stop-t2_start), " sec")
            
        return out


class NotImplementedToExpectedFailure:
    def _is_flakey(self):
        return self._testMethodName in getattr(self, "is_flakey", [])

    def _is_not_implemented(self):
        '''
        A test is expected to fail if:
          (a) Its name can be found in the test case's 'not_implemented' list
          (b) Its name can be found in the test case's 'is_flakey' list
          (c) Its name can be found in the test case's
              'not_implemented_versions' dictionary _and_ the current
              python version is in the dict entry's list
        :return: True if test is expected to fail
        '''
        method_name = self._testMethodName
        if method_name in getattr(self, 'not_implemented', []):
            return True

        if self._is_flakey():
            # -- Flakey tests sometimes fail, sometimes pass
            return True

        not_implemented_versions = getattr(self, 'not_implemented_versions', {})
        if method_name in not_implemented_versions:
            py_version = float("%s.%s" % (sys.version_info.major, sys.version_info.minor))
            if py_version in not_implemented_versions[method_name]:
                return True

        return False

    def run(self, result=None):
        # Override the run method to inject the "expectingFailure" marker
        # when the test case runs.
        if self._is_not_implemented():
            # Mark 'expecting failure' on class. It will only be applicable
            # for this specific run.

            # -- Save the original test_method _before_ we
            # overwrite it with wrapper
            test_method = getattr(self, self._testMethodName)

            def wrapper(*args, **kwargs):
                if self._is_flakey():
                    raise Exception("Flakey test that sometimes fails and sometimes passes")
                return test_method(*args, **kwargs)

            wrapper.__unittest_expecting_failure__ = True
            setattr(self, self._testMethodName, wrapper)
        return super().run(result=result)


SAMPLE_DATA = {
    'obj': [
            'object',
            'object()',
        ],
    'bool': [
            'True',
            'False',
        ],
    'bytearray': [
            'bytearray()',
            'bytearray(1)',
            'bytearray([1, 2, 3])',
            # 'bytearray(b"hello world")',
        ],
    'bytes': [
            'b""',
            'b"This is another string of bytes"',
            'b"One arg: %s"',
            'b"Three args: %s | %s | %s"',
        ],
    'class': [
            'type(1)',
            'type("a")',
            'type(object())',
            'type("MyClass", (object,), {})',
            # This datatype is here and must be here in order to "clear" the state set by type("MyClass", (object,), {})
            'type("object", (object,), {})'
        ],
    'complex': [
            '1j',
            '3.14159265j',
            '1+2j',
            '3-4j',
            '-5j',
        ],
    'dict': [
            '{}',
            '{"a": 1, "c": 2.3456, "d": "another"}',
        ],
    'float': [
            '2.3456',
            '0.0',
            '-3.14159',
            '-4.81756',
            '5.5',
            '-3.5',
            '4.5',
            '-4.5',
        ],
    'frozenset': [
            'frozenset()',
            'frozenset({"on","to","an"})',
            'frozenset({"one","two","six"})',
            'frozenset({1, 2.3456, 7})',
        ],
    'int': [
            '0',
            '-5',
            '-3',
            '3',
            '1',
        ],
    'list': [
            '[]',
            '[3, 4, 5]',
            '[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]',
            '["a","b","c"]',
        ],
    'range': [
            'range(0)',
            'range(5)',
            'range(2, 7)',
            'range(2, 7, 2)',
            'range(7, 2, -1)',
            'range(7, 2, -2)',
        ],
    'set': [
            'set()',
            'set({"on","to","an"})',
            'set({"one","two","six"})',
            'set({1, 2.3456, 7})',
        ],
    'slice': [
            'slice(0)',
            'slice(5)',
            'slice(2, 7)',
            'slice(2, 7, 2)',
            'slice(7, 2, -1)',
            'slice(7, 2, -2)',
        ],
    'str': [
            '""',
            '"3"',
            '"This is another string"',
            '"Mÿ hôvèrçràft îß fûłl öf éêlś"',
            '"One arg: %s"',
            '"Three args: %s | %s | %s"',
        ],
    'tuple': [
            '(1,)',
            '(1, 2)',
            '(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)',
            '(3, 1.2, True, )',
            '(1, 2.3456, "another")',
        ],
    'None': [
            'None',
        ],
    'NotImplemented': [
            'NotImplemented',
        ],
}


def _string_substitutions(string):
    delims = (string[0], string[-1])
    string = string[1:-1]
    elems = [elem for elem in string.split(", ")]
    perms = permutations(elems)
    return [(delims[0] + ", ".join(list(perm)) + delims[1]) for perm in perms]


SAMPLE_SUBSTITUTIONS = {
    # Normalize set ordering
    "{1, 2.3456, 7}": _string_substitutions("{1, 2.3456, 7}"),
    "{1, 2.3456, 'another'}": _string_substitutions("{1, 2.3456, 'another'}"),
    "{'an', 'to', 'on'}": _string_substitutions("{'an', 'to', 'on'}"),
    "{'one', 'two', 'six'}": _string_substitutions("{'one', 'two', 'six'}"),
    "{'a', 'b', 'c'}": _string_substitutions("{'a', 'b', 'c'}"),


    # Normalize list ordering
    "[1, 2.3456, 7]": _string_substitutions("[1, 2.3456, 7]"),
    "[1, 2.3456, 'another']": _string_substitutions("[1, 2.3456, 'another']"),
    "['a', 'b', 'c']": _string_substitutions("['a', 'b', 'c']"),
    "['an', 'to', 'on']": _string_substitutions("['an', 'to', 'on']"),

    "['one', 'two', 'six']": _string_substitutions("['one', 'two', 'six']"),
    # Normalize dictionary ordering
    "{'a': 1, 'c': 2.3456, 'd': 7}": _string_substitutions(
                                              "{'a': 1, 'c': 2.3456, 'd': 7}"),
    "{'a': 1, 'c': 2.3456, 'd': 'another'}": _string_substitutions(
                                      "{'a': 1, 'c': 2.3456, 'd': 'another'}"),
    "{'a': 'n', 't': 'o', 'o': 'n'}": _string_substitutions(
                                             "{'a': 'n', 't': 'o', 'o': 'n'}"),
    "{'to', 'two', 'one', 'an', 'on', 'six'}": _string_substitutions(
                                    "{'to', 'two', 'one', 'an', 'on', 'six'}"),
    "{1, 2.3456, 7, 'one', 'six', 'two'}": _string_substitutions(
                                        "{1, 2.3456, 7, 'one', 'six', 'two'}"),
    "{1, 2.3456, 7, 'on', 'an', 'to'}": _string_substitutions(
                                        "{1, 2.3456, 7, 'on', 'an', 'to'}"),
    "{1, 2.3456, 7, 1, 2.3456, 7}": _string_substitutions(
                                        "{1, 2.3456, 7, 1, 2.3456, 7}"),
    # Normalize precision error
    "-3.14159": ["-3.1415900000000008"],
}


def _unary_test(test_name, operation):
    def func(self):
        self.assertUnaryOperation(
            x_values=SAMPLE_DATA[self.data_type],
            operation=operation,
            format=self.format,
            substitutions=getattr(self, 'substitutions', SAMPLE_SUBSTITUTIONS)
        )
    return func


class UnaryOperationTestCase(NotImplementedToExpectedFailure):
    format = ''

    def assertUnaryOperation(self, x_values, operation, format, substitutions):
        self.assertCodeExecution(
            '##################################################\n'.join(
                adjust("""
                    try:
                        print('>>> x = %(x)s')
                        print('>>> %(format)s%(operation)sx')
                        x = %(x)s
                        print(%(format)s%(operation)sx)
                    except Exception as e:
                        print(type(e), ':', e)
                    print()
                    """ % {
                        'x': x,
                        'operation': operation,
                        'format': format,
                    }
                )
                for x in x_values
            ),
            "Error running %s" % operation,
            substitutions=substitutions,
            run_in_function=False,
        )

    test_unary_positive = _unary_test('test_unary_positive', '+')
    test_unary_negative = _unary_test('test_unary_negative', '-')
    test_unary_not = _unary_test('test_unary_not', 'not ')
    test_unary_invert = _unary_test('test_unary_invert', '~')


def _binary_test(test_name, operation, examples):
    def func(self):
        self.assertBinaryOperation(
            x_values=SAMPLE_DATA[self.data_type],
            y_values=examples,
            operation=operation,
            format=self.format,
            substitutions=getattr(self, 'substitutions', SAMPLE_SUBSTITUTIONS)
        )
    return func


class BinaryOperationTestCase(NotImplementedToExpectedFailure):
    format = ''
    y = 3

    def assertBinaryOperation(self, x_values, y_values, operation, format, substitutions):
        data = []
        for x in x_values:
            for y in y_values:
                data.append((x, y))

        self.assertCodeExecution(
            '##################################################\n'.join(
                adjust("""
                    try:
                        print('>>> x = %(x)s')
                        print('>>> y = %(y)s')
                        print('>>> %(format)s%(operation)s')
                        x = %(x)s
                        y = %(y)s
                        print(%(format)s%(operation)s)
                    except Exception as e:
                        print(type(e), ':', e)
                    print()
                    """ % {
                        'x': x,
                        'y': y,
                        'operation': operation,
                        'format': format,
                    }
                )
                for x, y in data
            ),
            "Error running %s" % operation,
            substitutions=substitutions,
            run_in_function=False,
        )

    for datatype, examples in SAMPLE_DATA.items():
        vars()['test_add_%s' % datatype] = _binary_test('test_add_%s' % datatype, 'x + y', examples)
        vars()['test_subtract_%s' % datatype] = _binary_test('test_subtract_%s' % datatype, 'x - y', examples)
        vars()['test_multiply_%s' % datatype] = _binary_test('test_multiply_%s' % datatype, 'x * y', examples)
        vars()['test_floor_divide_%s' % datatype] = _binary_test('test_floor_divide_%s' % datatype, 'x // y', examples)
        vars()['test_true_divide_%s' % datatype] = _binary_test('test_true_divide_%s' % datatype, 'x / y', examples)
        vars()['test_modulo_%s' % datatype] = _binary_test('test_modulo_%s' % datatype, 'x % y', examples)
        vars()['test_power_%s' % datatype] = _binary_test('test_power_%s' % datatype, 'x ** y', examples)
        vars()['test_subscr_%s' % datatype] = _binary_test('test_subscr_%s' % datatype, 'x[y]', examples)
        vars()['test_lshift_%s' % datatype] = _binary_test('test_lshift_%s' % datatype, 'x << y', examples)
        vars()['test_rshift_%s' % datatype] = _binary_test('test_rshift_%s' % datatype, 'x >> y', examples)
        vars()['test_and_%s' % datatype] = _binary_test('test_and_%s' % datatype, 'x & y', examples)
        vars()['test_xor_%s' % datatype] = _binary_test('test_xor_%s' % datatype, 'x ^ y', examples)
        vars()['test_or_%s' % datatype] = _binary_test('test_or_%s' % datatype, 'x | y', examples)

        vars()['test_lt_%s' % datatype] = _binary_test('test_lt_%s' % datatype, 'x < y', examples)
        vars()['test_le_%s' % datatype] = _binary_test('test_le_%s' % datatype, 'x <= y', examples)
        vars()['test_gt_%s' % datatype] = _binary_test('test_gt_%s' % datatype, 'x > y', examples)
        vars()['test_ge_%s' % datatype] = _binary_test('test_ge_%s' % datatype, 'x >= y', examples)
        vars()['test_eq_%s' % datatype] = _binary_test('test_eq_%s' % datatype, 'x == y', examples)
        vars()['test_ne_%s' % datatype] = _binary_test('test_ne_%s' % datatype, 'x != y', examples)

        vars()['test_direct_lt_%s' % datatype] = _binary_test('test_lt_%s' % datatype, 'x.__lt__(y)', examples)
        vars()['test_direct_le_%s' % datatype] = _binary_test('test_le_%s' % datatype, 'x.__le__(y)', examples)
        vars()['test_direct_gt_%s' % datatype] = _binary_test('test_gt_%s' % datatype, 'x.__gt__(y)', examples)
        vars()['test_direct_ge_%s' % datatype] = _binary_test('test_ge_%s' % datatype, 'x.__ge__(y)', examples)
        vars()['test_direct_eq_%s' % datatype] = _binary_test('test_eq_%s' % datatype, 'x.__eq__(y)', examples)
        vars()['test_direct_ne_%s' % datatype] = _binary_test('test_ne_%s' % datatype, 'x.__ne__(y)', examples)


def _inplace_test(test_name, operation, examples):
    def func(self):
        self.assertInplaceOperation(
            x_values=SAMPLE_DATA[self.data_type],
            y_values=examples,
            operation=operation,
            format=self.format,
            substitutions=getattr(self, 'substitutions', SAMPLE_SUBSTITUTIONS)
        )
    return func


class InplaceOperationTestCase(NotImplementedToExpectedFailure):
    format = ''
    y = 3

    def assertInplaceOperation(self, x_values, y_values, operation, format, substitutions):
        data = []
        for x in x_values:
            for y in y_values:
                data.append((x, y))

        self.assertCodeExecution(
            '##################################################\n'.join(
                adjust("""
                    try:
                        print('>>> x = %(x)s')
                        print('>>> y = %(y)s')
                        print('>>> %(operation)s')
                        print('>>> %(format)sx')
                        x = %(x)s
                        y = %(y)s
                        %(operation)s
                        print(%(format)sx)
                    except Exception as e:
                        print(type(e), ':', e)
                    print()
                    """ % {
                        'x': x,
                        'y': y,
                        'operation': operation,
                        'format': format,
                    }
                )
                for x, y in data
            ),
            "Error running %s" % operation,
            substitutions=substitutions,
            run_in_function=False,
        )

    for datatype, examples in SAMPLE_DATA.items():
        vars()['test_add_%s' % datatype] = _inplace_test(
            'test_add_%s' % datatype,
            'x += y',
            examples
        )
        vars()['test_subtract_%s' % datatype] = _inplace_test(
            'test_subtract_%s' % datatype,
            'x -= y',
            examples
        )
        vars()['test_multiply_%s' % datatype] = _inplace_test(
            'test_multiply_%s' % datatype,
            'x *= y',
            examples
        )
        vars()['test_floor_divide_%s' % datatype] = _inplace_test(
            'test_floor_divide_%s' % datatype,
            'x //= y',
            examples
        )
        vars()['test_true_divide_%s' % datatype] = _inplace_test(
            'test_true_divide_%s' % datatype,
            'x /= y',
            examples
        )
        vars()['test_modulo_%s' % datatype] = _inplace_test(
            'test_modulo_%s' % datatype,
            'x %= y',
            examples
        )
        vars()['test_power_%s' % datatype] = _inplace_test(
            'test_power_%s' % datatype,
            'x **= y',
            examples
        )
        vars()['test_lshift_%s' % datatype] = _inplace_test(
            'test_lshift_%s' % datatype,
            'x <<= y',
            examples
        )
        vars()['test_rshift_%s' % datatype] = _inplace_test(
            'test_rshift_%s' % datatype,
            'x >>= y',
            examples
        )
        vars()['test_and_%s' % datatype] = _inplace_test(
            'test_and_%s' % datatype,
            'x &= y',
            examples
        )
        vars()['test_xor_%s' % datatype] = _inplace_test(
            'test_xor_%s' % datatype,
            'x ^= y',
            examples
        )
        vars()['test_or_%s' % datatype] = _inplace_test(
            'test_or_%s' % datatype,
            'x |= y',
            examples
        )


def _builtin_test(test_name, operation, examples):
    def func(self):
        self.assertBuiltinFunction(
            x_values=examples,
            f_values=self.functions,
            operation=operation,
            format=self.format,
            substitutions=getattr(self, 'substitutions', SAMPLE_SUBSTITUTIONS)
        )
    return func


class BuiltinFunctionTestCase(NotImplementedToExpectedFailure):
    format = ''

    def assertBuiltinFunction(self, f_values, x_values, operation, format, substitutions):
        data = []
        for f in f_values:
            for x in x_values:
                data.append((f, x))

        self.assertCodeExecution(
            '##################################################\n'.join(
                adjust("""
                    try:
                        print('>>> f = %(f)s')
                        print('>>> x = %(x)s')
                        print('>>> %(format)s%(operation)s')
                        f = %(f)s
                        x = %(x)s
                        print(%(format)s%(operation)s)
                    except Exception as e:
                        print(type(e), ':', e)
                    print()
                    """ % {
                        'f': f,
                        'x': x,
                        'operation': operation,
                        'format': format,
                    }
                )
                for f, x in data
            ),
            "Error running %s" % operation,
            substitutions=substitutions,
            run_in_function=False,
        )

    for datatype, examples in SAMPLE_DATA.items():
        vars()['test_%s' % datatype] = _builtin_test('test_%s' % datatype, 'f(x)', examples)


def _builtin_twoarg_test(test_name, operation, examples1, examples2):
    def func(self):
        self.assertBuiltinTwoargFunction(
            f_values=self.functions,
            x_values=examples1,
            y_values=examples2,
            operation=operation,
            format=self.format,
            substitutions=getattr(self, 'substitutions', SAMPLE_SUBSTITUTIONS)
        )
    return func


class BuiltinTwoargFunctionTestCase(NotImplementedToExpectedFailure):
    format = ''

    def assertBuiltinTwoargFunction(self, f_values, x_values, y_values, operation, format, substitutions):
        data = []
        for f in f_values:
            for x in x_values:
                for y in y_values:
                    data.append((f, x, y))

        self.assertCodeExecution(
            '##################################################\n'.join(
                adjust("""
                    try:
                        print('>>> f = %(f)s')
                        print('>>> x = %(x)s')
                        print('>>> y = %(y)s')
                        print('>>> %(format)s%(operation)s')
                        f = %(f)s
                        x = %(x)s
                        y = %(y)s
                        print(%(format)s%(operation)s)
                    except Exception as e:
                        print(type(e), ':', e)
                    print()
                    """ % {
                        'f': f,
                        'x': x,
                        'y': y,
                        'operation': operation,
                        'format': format,
                    }
                )
                for f, x, y in data
            ),
            "Error running %s" % operation,
            substitutions=substitutions,
            run_in_function=False,
        )

    for datatype1, examples1 in SAMPLE_DATA.items():
        for datatype2, examples2 in SAMPLE_DATA.items():
            vars()['test_%s_%s' % (datatype1, datatype2)] = _builtin_twoarg_test(
                'test_%s_%s' % (datatype1, datatype2),
                'f(x, y)', examples1, examples2
            )



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class LenTests(TranspileTestCase):
    pass


class BuiltinLenFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["len"]

    not_implemented = []



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class IterTests(TranspileTestCase):
    pass


class BuiltinIterFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["iter"]

    not_implemented = [
        'test_bytearray',
        'test_bytes',
        'test_str',
    ]



from .. utils import TranspileTestCase, BuiltinFunctionTestCase


class CallableTests(TranspileTestCase):
    pass


class BuiltinCallableFunctionTests(BuiltinFunctionTestCase, TranspileTestCase):
    functions = ["callable"]

    not_implemented = [
    ]



from ..utils import TranspileTestCase


class ScopeTests(TranspileTestCase):
    def test_simple(self):
        # This test will run both in global scope, and in a
        # function.
        self.assertCodeExecution("""
            x = 1
            print('x =', x)
            """)

    def test_local_scope(self):
        self.assertCodeExecution("""
            x = 1

            def foo():
                print("1: x =", x)
                x = 2
                print("2: x =", x)

            print("3: x =", x)
            try:
                foo()
            except UnboundLocalError as err:
                print(err)
            print("4: x =", x)
            """, run_in_function=False)

    def test_global_scope(self):
        self.assertCodeExecution("""
            x = 1

            def foo():
                global x
                print("1: x =", x)
                x = 2
                print("2: x =", x)

            print("3: x =", x)
            foo()
            print("4: x =", x)

            """, run_in_function=False)

    def test_class_scope(self):
        self.assertCodeExecution("""
            x = 1

            class Foo:
                def __init__(self, y):
                    print("1: y =", y)
                    x = 2
                    print("2: x =", x)
                    self.z = y

                w = x
                v = abs(x)
                x = 4

            print("3: x =", x)
            f = Foo(3)
            print("3: Foo.v =", Foo.v)
            print("3: f.v =", f.v)
            print("3: Foo.w =", Foo.w)
            print("3: f.w =", f.w)
            print("3: x =", x)
            try:
                print("3: f.x =", f.x)
            except AttributeError:
                pass
            try:
                print("3: f.y =", f.y)
            except AttributeError:
                pass
            print("3: f.z =", f.z)
            """, run_in_function=False)



from unittest import expectedFailure

from ..utils import TranspileTestCase


class TryExceptTests(TranspileTestCase):
    def test_try_except(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except:
                print("Got an error")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except:
                print("Got an error")
            print('Done.')
            """)

        # Test expression statement
        self.assertCodeExecution("""
            try:
                [1, 2, 3]
                "abc" == "def"
            except:
                print("Got an error")
            print('Done.')
            """)

    def test_try_except_unnamed(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got an error")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got an error")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got an error")
            print('Done.')
            """, exits_early=True)

    def test_try_except_named(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError as e:
                print("Got a", e)
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError as e:
                print("Got a", e)
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError as e:
                print("Got a", e)
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        # Caught first exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        # Caught second exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError:
                print("Got an AttributeError")
            except NameError:
                print("Got a NameError")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_named(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        # Caught first exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """)

        # Caught second exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError as e:
                print("Got an AttributeError", e)
            except NameError as e:
                print("Got a NameError", e)
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_match_except_unnamed(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except (NameError, TypeError):
                print("Got a NameError or TypeError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        # Caught first exception, first handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except (NameError, TypeError):
                print("Got a NameError or TypeError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        # Caught second exception, first handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except (TypeError, NameError):
                print("Got a TypeError or NameError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        # Caught first exception, second handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError:
                print("Got an AttributeError")
            except (NameError, TypeError):
                print("Got a NameError or TypeError")
            print('Done.')
            """)

        # Caught second exception, second handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError:
                print("Got an AttributeError")
            except (TypeError, NameError):
                print("Got a TypeError or NameError")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except (NameError, TypeError):
                print("Got a NameError or TypeError")
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_match_except_named(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except (NameError, TypeError) as e:
                print("Got a NameError or TypeError", e)
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """)

        # Caught first exception, first handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except (NameError, TypeError) as e:
                print("Got a NameError or TypeError", e)
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """)

        # Caught second exception, first handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except (TypeError, NameError) as e:
                print("Got a TypeError or NameError", e)
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """)

        # Caught first exception, second handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError as e:
                print("Got an AttributeError", e)
            except (NameError, TypeError) as e:
                print("Got a NameError or TypeError", e)
            print('Done.')
            """)

        # Caught second exception, second handler
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError as e:
                print("Got an AttributeError", e)
            except (TypeError, NameError) as e:
                print("Got a TypeError or NameError", e)
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except (NameError, TypeError) as e:
                print("Got a NameError or TypeError", e)
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_mixed1(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_mixed2(self):
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """)

        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError:
                print("Got an AttributeError")
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_mixed3(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError:
                print("Got an AttributeError")
            except:
                print("Got an anonymous error")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError:
                print("Got an AttributeError")
            except:
                print("Got an anonymous error")
            print('Done.')
            """)

        # Caught as bucket case
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            except AttributeError:
                print("Got an AttributeError")
            except:
                print("Got an anonymous error")
            print('Done.')
            """)

    @expectedFailure
    def test_nested_try_catch_same_error(self):
        self.assertCodeExecution("""
            try:
                try:
                    raise TypeError
                except TypeError as e:
                    print("handled by first except")

                try:
                    raise TypeError
                except TypeError as e:
                    print("handled by second except")
            except TypeError as e:
                print("handled by outer except")
            """)

    def test_try_catch_in_loop(self):
        self.assertCodeExecution("""
            for x in range(2):
                try:
                    print('success', next(iter([])))
                except Exception:
                    print('whoops')
            """)

    def test_try_nested(self):
        self.assertCodeExecution("""
            try:
                print('outer try')
                try:
                    print('inner try')
                except:
                    print('inner except')
            except:
                print('outer except')
        """)

    def test_try_if_else_in_try(self):
        self.assertCodeExecution("""
            try:
                if 1 == 2:
                    print('1 == 2!?')
                else:
                    print('1 != 2')
            except:
                print('except')
        """)


class TryExceptFinallyTests(TranspileTestCase):
    def test_try_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Uncaught Exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            finally:
                print("Do final cleanup")
            print('Done.')
            """, exits_early=True)

    def test_try_except_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except:
                print("Got an error")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught Exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except:
                print("Got an error")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

    def test_try_except_unnamed_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got an error")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got an error")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got an error")
            finally:
                print("Do final cleanup")
            print('Done.')
            """, exits_early=True)

    def test_try_except_named_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError as e:
                print("Got a NameError", e)
            finally:
                print("Do final cleanup")
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught first exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught second exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except AttributeError as e:
                print("Got an AttributeError", e)
            except NameError:
                print("Got a NameError")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            finally:
                print("Do final cleanup")
            print('Done.')
            """, exits_early=True)


class TryExceptElseTests(TranspileTestCase):
    def test_try_except_else(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except:
                print("Got an error")
            else:
                print("Do else handling")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except:
                print("Got an error")
            else:
                print("Do else handling")
            print('Done.')
            """)

    def test_try_except_unnamed_else(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got an error")
            else:
                print("Do else handling")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got an error")
            else:
                print("Do else handling")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got an error")
            else:
                print("Do else handling")
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_else(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            else:
                print("Do else handling")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            else:
                print("Do else handling")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            else:
                print("Do else handling")
            print('Done.')
            """, exits_early=True)


class TryExceptElseFinallyTests(TranspileTestCase):
    def test_try_except_else_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except:
                print("Got an error")
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except:
                print("Got an error")
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

    def test_try_except_unnamed_else_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj = 3
                print('OK')
            except NameError:
                print("Got an error")
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got an error")
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got an error")
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """, exits_early=True)

    def test_try_multiple_except_else_finally(self):
        # No exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Caught exception
        self.assertCodeExecution("""
            try:
                obj.no_such_attribute
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """)

        # Uncaught exception
        self.assertCodeExecution("""
            try:
                obj = int('asdf')
                print('OK')
            except NameError:
                print("Got a NameError")
            except AttributeError as e:
                print("Got an AttributeError", e)
            else:
                print("Do else handling")
            finally:
                print("Do final cleanup")
            print('Done.')
            """, exits_early=True)
